#ifdef _WINDOWS
#define WINDOWS
#endif

#include "ac_plugin.h"
#include "Undoables.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef WINDOWS
#include <windows.h>
#else
#include <time.h>
#endif

static unsigned int ddhrandseed = 0;

//---------------------------------------------------------------------------------------
// Get a random seed number
static unsigned int ddhGetRandomSeed()
{
#ifdef WINDOWS
   return GetTickCount();
#else
   ddhrandseed += time( NULL ) / 2;
   return ddhrandseed;
#endif
}
//---------------------------------------------------------------------------------------
// Append list with object's vertices
static void ddhwave_add_object_recurse( ACObject *o, List **v )
{
   list_append_list_unique( v, ac_object_get_vertexlist( o ));
   List *c = ac_object_get_childrenlist( o );
   for( List *cl = c; cl != NULL; cl = cl->next )
   {
      ACObject *o = ( ACObject * )( cl->data );
      ddhwave_add_object_recurse( o, v );
   }
}
//---------------------------------------------------------------------------------------
// Applies a waveform to selected vertices
static void ddhwave_go( int argc, const char* argv[] )
{

   if( argc < 12 )
   {
      message_dialog( "Insufficient parameters passed to wave utility" );
      return;
   }

   bool random1 = false;
   bool random2 = false;
   unsigned int seed = 0;

   // Get params
   int i = 0;
   char axis_freq = tolower( argv[++i][0] );
   char axis_amp1 = tolower( argv[++i][0] );
   char axis_amp2 = tolower( argv[++i][0] );

   if( axis_freq == axis_amp1 || axis_freq == axis_amp2 )
   {
      message_dialog( "Wave3d : cannot apply frequency and amplitude to same axis" );
      return;
   }

   else if( axis_freq != 'x' && axis_freq != 'y' && axis_freq != 'z' )
   {
      message_dialog( "Wave3d : cannot interpret frequency axis %c", axis_freq );
      return;
   }

   else if( axis_amp1 == axis_amp2 )
   {
      message_dialog( "Wave3d : the two amplitude axis should differ, both = '%c'", axis_amp1 );
      return;
   }

   else if( axis_amp1 != 'x' && axis_amp1 != 'y' && axis_amp1 != 'z' && axis_amp1 != 'n' )
   {
      message_dialog( "Wave3d : cannot interpret amplitude axis %c", axis_amp1 );
      return;
   }

   else if( axis_amp2 != 'x' && axis_amp2 != 'y' && axis_amp2 != 'z' && axis_amp2 != 'n')
   {
      message_dialog( "Wave3d : cannot interpret amplitude axis %c", axis_amp2 );
      return;
   }


   float frequency, amplitude1, amplitude2, decay1, decay2, offset1, offset2;
   bool additive, frequnits;

   // Frequency
   if( sscanf( argv[++i], "%g", &frequency ) != 1 )
   {
      message_dialog( "Wave3d : unable to parse %s for frequency", argv[i] );
      return;
   }


   // Amplitude 1
   if( sscanf( argv[++i], "%g", &amplitude1 ) != 1 )
   {
      message_dialog( "Wave3d : unable to parse %s for amplitude 1", argv[i] );
      return;
   }

   // Amplitude 2
   if( sscanf( argv[++i], "%g", &amplitude2 ) != 1 )
   {
      message_dialog( "Wave3d : unable to parse %s for amplitude 2", argv[i] );
      return;
   }

   // Decay 1
   if( sscanf( argv[++i], "%g", &decay1 ) != 1 )
   {
      message_dialog( "Wave3d : unable to parse %s for decay 1", argv[i] );
      return;
   }

   // Decay 2
   if( sscanf( argv[++i], "%g", &decay2 ) != 1 )
   {
      message_dialog( "Wave3d : unable to parse %s for decay 2", argv[i] );
      return;
   }


   // Offset 1
   if( strcmp( argv[++i], "random" ) == 0 )
   {
      seed = ddhGetRandomSeed();
      srand( seed );
      offset1 = (( float )( rand() % 100 )) * 0.01f;
      random1 = true;
   }
   else
   {
      if( sscanf( argv[i], "%g", &offset1 ) != 1 )
      {
         message_dialog( "Wave3d : unable to parse %s for offset 1", argv[i] );
         return;
      }
      offset1 *= 0.01f;
   }

   // Offset 2
   if( strcmp( argv[++i], "random" ) == 0 )
   {
      seed = ddhGetRandomSeed();
      srand( seed );
      offset2 = (( float )( rand() % 100 )) * 0.01f;
      random2 = true;
   }
   else
   {
      if( sscanf( argv[i], "%g", &offset2 ) != 1 )
      {
         message_dialog( "Wave3d : unable to parse %s for offset 2", argv[i] );
         return;
      }
      offset2 *= 0.01f;
   }


   additive = !( argv[++i][0] == '0' );
   frequnits = !( argv[++i][0] == '0' );

   debugf( "Wave3D:\nFrequency axis = %c\nAmplitude axis 1 = %c\nAmplitude axis 2 = %c\nFrequency = %g\nAmplitude 1 = %g\nAmplitude 2 = %g\nDecay 1 = %g\nDecay 2 = %g\nOffset 1 = %g\nOffset 2 = %g\nAdditive = %s\nFreq units = %s\n",
      axis_freq,
      axis_amp1, axis_amp2,
      frequency,
      amplitude1, amplitude2,
      decay1, decay2,
      offset1, offset2,
      additive  ? "true" : "false",
      frequnits ? "true" : "false" );

  decay1 *= 0.01f;
  decay2 *= 0.01f;

   // Changed vertex list
   List *vertlist = NULL;
   switch( ac_get_select_mode() )
   {
   case SELECT_SURFACE:
      {
         List *l = ac_selection_get_surfaces();
         vertlist = ac_surfacelist_get_vertices( l );
         list_free( &l );
      }
      break;
   case SELECT_VERTEX:
      vertlist = ac_selection_get_vertices();
      break;
   default:
      {
         List *l = ac_selection_get_objects();
         for( List *lp = l; lp != NULL; lp = lp->next )
         {
            ACObject *o = ( ACObject * )( lp->data );
            ddhwave_add_object_recurse( o, &vertlist );
         }
         list_free( &l );
      }
      break;
   }


   if( vertlist == NULL )
   {
      message_dialog( "Wave3d : no vertices selected" );
      return;
   }

   // Add undoable
   add_undoable_vertex_positions( "Wave3d", vertlist );

   float ampbase1 = 0.0f,  ampbase2 = 0.0f;
   float freqsiz = 0.0f,  freqmin = 0.0f,  freqmax = 0.0f;
   Vertex *vminaxis = NULL;
   Vertex *vmaxaxis = NULL;
   float ampbasefreq = 0.0f;
   float ampmaxval = 0.0f;

   // Grab the min frequency vertex, use as the base vertex amplitude value
   for( List *lv = vertlist; lv != NULL; lv = lv->next )
   {
      Vertex *v = ( Vertex * )( lv->data );
      if( vminaxis == NULL )
      {
         vminaxis = v;
         vmaxaxis = v;
         switch( axis_freq )
         {
         case 'x':
            ampbasefreq = v->x;
            ampmaxval = v->x;
            break;
         case 'y':
            ampbasefreq = v->y;
            ampmaxval = v->y;
            break;
         case 'z':
            ampbasefreq = v->z;
            ampmaxval = v->z;
            break;
         }
         // Do not continue - otherwise, single vertex runs will screw up
         //continue;
      }
      switch( axis_freq )
      {
      case 'x':
         if( v->x < vminaxis->x )
         {
            vminaxis = v;
            ampbasefreq = v->x;
            freqmin = v->x;
         }
         if( v->x > vmaxaxis->x )
         {
            vmaxaxis = v;
            ampmaxval = v->x;
            freqmax = v->x;
         }
         break;
      case 'y':
         if( v->y < vminaxis->y )
         {
            vminaxis = v;
            ampbasefreq = v->y;
            freqmin = v->y;
         }
         if( v->y > vmaxaxis->y )
         {
            vmaxaxis = v;
            ampmaxval = v->y;
            freqmax = v->y;
         }
         break;
      case 'z':
         if( v->z < vminaxis->z )
         {
            vminaxis = v;
            ampbasefreq = v->z;
            freqmin = v->z;
         }
         if( v->z > vmaxaxis->z )
         {
            vmaxaxis = v;
            ampmaxval = v->z;
            freqmax = v->z;
         }
         break;
      }
   }

   freqsiz = freqmax - freqmin;

   switch( axis_amp1 )
   {
   case 'x':
      ampbase1 = vminaxis->x;
      break;
   case 'y':
      ampbase1 = vminaxis->y;
      break;
   case 'z':
      ampbase1 = vminaxis->z;
      break;
   }

   switch( axis_amp2 )
   {
   case 'x':
      ampbase2 = vminaxis->x;
      break;
   case 'y':
      ampbase2 = vminaxis->y;
      break;
   case 'z':
      ampbase2 = vminaxis->z;
      break;
   }

   // If frequency is not expressed in units, grab the frequency based on the number of waves in the total length
   if( !frequnits )
   {
      offset1 *= frequency;
      offset2 *= frequency;
      if( ampmaxval - ampbasefreq != 0.0f )
      {
         frequency /= ( ampmaxval - ampbasefreq );
      }
      else
      {
         frequency = 0;
      }
   }

   // Transform all verts
   float f, *a1, *a2;
   float ignore = 0.0f;
   float ampof = 0.0f;
   ACObject *thisobj;
   ACObject *lastobj = NULL;
   bool subdivided = false;

   if( axis_amp1 == 'n' ) {amplitude1 = 0.0f; }
   if( axis_amp2 == 'n' ) {amplitude2 = 0.0f; }

   for( List *lv = vertlist; lv != NULL; lv = lv->next )
   {

      Vertex *v = ( Vertex * )( lv->data );
      thisobj = object_of_vertex( v );

      switch( axis_freq )
      {
      case 'x':
         f = v->x - vminaxis->x;
         break;
      case 'y':
         f = v->y - vminaxis->y;
         break;
      case 'z':
         f = v->z - vminaxis->z;
         break;
      }

      switch( axis_amp1 )
      {
      case 'x':
         a1 = &( v->x );
         break;
      case 'y':
         a1 = &( v->y );
         break;
      case 'z':
         a1 = &( v->z );
         break;
      case 'n':
         a1 = &( ignore );
         break;

      default:
         message_dialog( "opos" );
      }

      switch( axis_amp2 )
      {
      case 'x':
         a2 = &( v->x );
         break;
      case 'y':
         a2 = &( v->y );
         break;
      case 'z':
         a2 = &( v->z );
         break;
      case 'n':
         a2 = &( ignore );
         break;
      default:
         message_dialog( "opos" );
      }



      // Change random offset for each object
      if( thisobj != lastobj )
      {
         subdivided = subdivided | ( ac_object_is_subdivided( thisobj ) != FALSE );
         if( random1 )
         {
            lastobj = thisobj;
            seed += 7;
            srand( seed );
            offset1 = (( float )( rand() % 100 )) * 0.01f;
            if( !frequnits )
            {
               offset1 *= frequency;
            }
         }
         if( random2 )
         {
            lastobj = thisobj;
            seed += 7;
            srand( seed );
            offset2 = (( float )( rand() % 100 )) * 0.01f;
            if( !frequnits )
            {
               offset2 *= frequency;
            }
         }

      }

      // The wave delta
      debugf( "\nWave3D:\nfreqsiz = %g f = %g Math = %g Freq units = %s\n",
         freqsiz,
         f,
         ( freqsiz - f ) / freqsiz,
         frequnits ? "true" : "false" );

      //float curdecay1 = decay1 * (( freqsiz - f ) / freqsiz );
      //float curdecay2 = decay2 * (( freqsiz - f ) / freqsiz );


      float curdecay1 = 1.0f - (decay1 * (1.0f - ( freqsiz - f ) / freqsiz));
      float curdecay2 = 1.0f - (decay2 * (1.0f - ( freqsiz - f ) / freqsiz));

      float w_sin = sin(((( f * frequency ) + offset1 ) * 360.0f ) * M_PI / 180.0f ) * amplitude1 * curdecay1;
      float w_cos = cos(((( f * frequency ) + offset2 ) * 360.0f ) * M_PI / 180.0f ) * amplitude2 * curdecay2;


      if( additive ) {
         *a1 += w_sin;
         *a2 += w_cos;
      } else {
         *a1 = ampbase1 + w_sin;
         *a2 = ampbase2 + w_cos;
      }

   }

   // Redraw/refresh
   if( subdivided )
   {
      tcl_command( "ac3d remake_subdivisions" );
   }
   find_bounding_box();
   redraw_all();
   display_status();

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      char *path = ( char * )strdup( d->plugin_dir_path );
#ifdef WINDOWS
      ac_unbackslash_string( path );
#endif

      tcl_command( "source \"%s/wave3d.tcl\"", path );
      ac_add_tools_menu_item( "Wave3D...", "ddhwave_dialog",
                              "Apply an axis-aligned waveform to the selected vertices" );
      ac_add_command_full( "ddhwave_go", ( void * )ddhwave_go,
                            12, "argv", "TODO: Show params here",
                           "Wave3D utility" );

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Wave3D - version 1.00 - Dennis Hawthorne and Eldon Brown" );
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Wave3D utility - Applies an axis-aligned waveform to selected vertices" );
}
//---------------------------------------------------------------------------------------

